@@ -443,6 +443,9 @@ class ParameterEstimator(PEMixins, Optimizer):
 
             expr *= weights[0]
             second_term = 0.0
+            print('These are the variances')
+            print(sigma_sq)
+            
             for t in m.meas_times:
                 second_term += sum((model.C[t, k] - model.Z[t, k]) ** 2 / sigma_sq[k] for k in list_components)
                 
