@@ -30,6 +30,7 @@ from kipet.core_methods.PyomoSimulator import PyomoSimulator
 from kipet.core_methods.TemplateBuilder import TemplateBuilder
 from kipet.core_methods.VarianceEstimator import VarianceEstimator
 # from kipet.common.component_expression import get_unit_model
+from kipet.common.interpolation import interpolate_trajectory
 from kipet.common.model_funs import step_fun
 from kipet.post_model_build.pyomo_model_tools import get_vars
 from kipet.dev_tools.display import Print
@@ -37,6 +38,7 @@ from kipet.post_model_build.scaling import scale_models
 from kipet.post_model_build.replacement import ParameterReplacer
 from kipet.mixins.TopLevelMixins import WavelengthSelectionMixins
 
+
 from kipet.top_level.data_component import (
     DataBlock, 
     DataSet,
@@ -65,9 +67,10 @@ from pyomo.environ import units as pyo_units
 from kipet.model_components.units_handler import convert_single_dimension
 
 __var = VariableNames()
-DEBUG=__var.DEBUG
+DEBUG=True #__var.DEBUG
 _print = Print(verbose=DEBUG)
 model_components = ['parameters', 'components', 'states', 'algebraics', 'constants', 'steps']
+version_number = '0.2.2'
 
 class ReactionModel(WavelengthSelectionMixins):
     
@@ -85,7 +88,9 @@ class ReactionModel(WavelengthSelectionMixins):
         
         self.name = kwargs.get('name', 'Model-1')
         self.model = None
+        self.s_model = None
         self.builder = TemplateBuilder()
+        self._template_populated = False
         self.components = ComponentBlock()   
         self.parameters = ParameterBlock()
         self.constants = ConstantBlock()
@@ -116,6 +121,7 @@ class ReactionModel(WavelengthSelectionMixins):
         self._var_to_initialize_from_trajectory = []
         self._default_time_unit = 'seconds'
         self.unit_base = kwargs.get('unit_base', None)
+        self.allow_optimization = False
         
         self._G_data = {'G_contribution': None, 'Z_in': dict(), 'St': dict()}
         self.__var = VariableNames()
@@ -177,6 +183,9 @@ class ReactionModel(WavelengthSelectionMixins):
         """
         if not hasattr(self, '_set_up_model'):
             self._make_set_up_model()
+
+        if hasattr(self._set_up_model, name):        
+            raise ValueError('A component with this name has already been defiffffned')
         
         setattr(self._set_up_model, f'{name}_indx', Set(initialize=[0]))
         sets = [getattr(self._set_up_model, f'{name}_indx')]*index
@@ -210,6 +219,9 @@ class ReactionModel(WavelengthSelectionMixins):
         value = kwargs.get('value', 1)
         kwargs['units_orig'] = units
         
+        comp_name = args[0]
+        # if comp_name in getattr(self, f'{name}s'):
+        #     print('repeat')
         par, con = self.Component(args[0], index, units=units)
         #print(f'#### in the add model: {con}')
         
@@ -490,6 +502,14 @@ class ReactionModel(WavelengthSelectionMixins):
     
     """ Expressions """
     
+    def add_reaction(self, name, expr, *args, **kwargs):
+        """Wrapper to add reactions without needing the kwargs, simplifies
+        the API somewhat. The default option for algebraics is False!
+        """
+        kwargs['is_reaction'] = True
+        return self.add_expression(name, expr, *args, **kwargs)
+        
+    
     def add_expression(self, name, expr, *args, **kwargs):
         """Add expressions (prev. algebraics) to the reaction model
         
@@ -642,85 +662,93 @@ class ReactionModel(WavelengthSelectionMixins):
         """Method handling all of the preparation for the TB
         
         """
-        with_data = kwargs.get('with_data', True)
         
-        # for model_component, obj in self.component_map.items():
+        if not self._template_populated:
+        
+            with_data = kwargs.get('with_data', True)
+                    
+            if len(self.states) > 0:
+                self.builder.add_model_element(self.states)
             
-        #     if len(obj) > 0:
-        #         self.builder.add_model_element(obj)
+            if len(self.algebraics) > 0:
+                self.builder.add_model_element(self.algebraics)
+            
+            if len(self.components) > 0:
+                self.builder.add_model_element(self.components)
+            else:
+                raise ValueError('The model has no components')
                 
-        if len(self.states) > 0:
-            self.builder.add_model_element(self.states)
-        
-        if len(self.algebraics) > 0:
-            self.builder.add_model_element(self.algebraics)
-        
-        if len(self.components) > 0:
-            self.builder.add_model_element(self.components)
-        else:
-            raise ValueError('The model has no components')
-            
-        if len(self.parameters) > 0:
-            self.builder.add_model_element(self.parameters)
-        else:
-            self.allow_optimization = False   
-        
-        if len(self.constants) > 0:
-            self.builder.add_model_element(self.constants)
-        
-        if with_data:
-            if len(self.datasets) > 0 or self.spectra is not None:
-                self.builder.input_data(self.datasets, self.spectra)
-                self.allow_optimization = True
-            elif len(self.datasets) == 0 and self.spectra is None:
-                self.allow_optimization = False
+            if len(self.parameters) > 0:
+                self.builder.add_model_element(self.parameters)
             else:
+                self.allow_optimization = False   
+            
+            if len(self.constants) > 0:
+                self.builder.add_model_element(self.constants)
+            
+            if with_data:
+                if len(self.datasets) > 0 or self.spectra is not None:
+                    self.builder.input_data(self.datasets, self.spectra)
+                    self.allow_optimization = True
+                elif len(self.datasets) == 0 and self.spectra is None:
+                    self.allow_optimization = False
+                else:
+                    pass
+                
+            # Add the ODEs
+            if len(self.odes_dict) != 0:
+                self._build_odes()
+                self.builder.set_odes_rule(self.odes)
+            elif 'odes_bypass' in kwargs and kwargs['odes_bypass']:
                 pass
-            
-        # Add the ODEs
-        if len(self.odes_dict) != 0:
-            self._build_odes()
-            self.builder.set_odes_rule(self.odes)
-        elif 'odes_bypass' in kwargs and kwargs['odes_bypass']:
-            pass
-        else:
-            raise ValueError('The model requires a set of ODEs')
-            
-        if len(self.algs_dict) != 0:
-            self._build_algs()
-            if isinstance(self.algs, dict):
-                self.builder.set_algebraics_rule(self.algs, asdict=True)
             else:
-                self.builder.set_algebraics_rule(self.algs)
+                raise ValueError('The model requires a set of ODEs')
+                
+            if len(self.algs_dict) != 0:
+                self._build_algs()
+                if isinstance(self.algs, dict):
+                    self.builder.set_algebraics_rule(self.algs, asdict=True)
+                else:
+                    self.builder.set_algebraics_rule(self.algs)
+                
+            if hasattr(self, 'custom_objective') and self.custom_objective is not None:
+                self.builder.set_objective_rule(self.custom_objective)
             
-        if hasattr(self, 'custom_objective') and self.custom_objective is not None:
-            self.builder.set_objective_rule(self.custom_objective)
-        
-        self.builder.set_parameter_scaling(self.settings.general.scale_parameters)
-        self.builder.add_state_variance(self.components.variances)
-        
-        # if self._has_step_or_dosing:
-        #     self.builder.add_dosing_var(self._number_of_steps)
-        
-        if self._has_dosing_points:
-            self._add_feed_times()
+            self.builder.set_parameter_scaling(self.settings.general.scale_parameters)
+            self.builder.add_state_variance(self.components.variances)
             
-        if hasattr(self, '_step_list') and len(self._step_list) > 0:
-            self.builder.add_step_vars(self._step_list)
+            # if self._has_step_or_dosing:
+            #     self.builder.add_dosing_var(self._number_of_steps)
             
-        # It seems this is repetitive - refactor
-        if hasattr(self, '_G_contribution') and self._G_contribution is not None:
-        #if self.settings.parameter_estimator.G_contribution is not None:
-            self._unwanted_G_initialization()
-            self.builder._G_contribution = self._G_contribution
+            if self._has_dosing_points:
+                self._add_feed_times()
+                
+            if hasattr(self, '_step_list') and len(self._step_list) > 0:
+                self.builder.add_step_vars(self._step_list)
+                
+            # It seems this is repetitive - refactor
+            if hasattr(self, '_G_contribution') and self._G_contribution is not None:
+            #if self.settings.parameter_estimator.G_contribution is not None:
+                self._unwanted_G_initialization()
+                self.builder._G_contribution = self._G_contribution
+            else:
+                self.builder._G_contribution = None
+            
+            self._template_populated = True
+        
         else:
-            self.builder._G_contribution = None
+            print('Template already populated')
+            
+        return None
+           
+            
+    def _get_model_times(self):
         
         start_time, end_time = None, None
         if self.settings.general.simulation_times is not None:
             #print(f'times are: {type(self.settings.general.simulation_times)}')
             start_time, end_time = self.settings.general.simulation_times
-       
+
         return start_time, end_time
     
     def _make_c_dict(self):
@@ -732,8 +760,6 @@ class ReactionModel(WavelengthSelectionMixins):
             for mc in model_components: 
                 if hasattr(self, f'{mc}') and len(getattr(self, f'{mc}')) > 0:
                     for comp in getattr(self, f'{mc}'):
-                        #print(comp)
-                        #print(comp.model_var, comp.pyomo_var)
                         self.c.update({comp.name: [comp.model_var, comp.pyomo_var]})
     
             if hasattr(self, 'steps'):
@@ -745,33 +771,34 @@ class ReactionModel(WavelengthSelectionMixins):
         and there is nothing returned.
 
         Args:
-            None
+            is_simulation (bool): Default False, tells the builder if a
+            simulation or optimization models is being built
 
         Returns:
-            None
+            pyomo_model (ConcreteModel): The resulting Pyomo model
 
         """
-        if hasattr(self, 'model'):
-            del self.model
-            
+        kwargs['is_simulation'] = kwargs.get('is_simulation', False)
         skip_non_abs = kwargs.pop('skip_non_abs', False)
-            
-        start_time, end_time = self.populate_template(*args, **kwargs)
+         
+        self.populate_template(*args, **kwargs)
         self._make_c_dict()
         setattr(self.builder, 'c_mod', self.c)
         
-        self.model = self.builder.create_pyomo_model(start_time, end_time)
+        start_time, end_time = self._get_model_times()
+        if self.model is None:
+            self.model = self.builder.create_pyomo_model(start_time, end_time, False)
+        pyomo_model = self.builder.create_pyomo_model(start_time, end_time, kwargs['is_simulation'])
         
-        non_abs_comp = self.components.get_match('absorbing', False)
+        if not kwargs['is_simulation']:
+            non_abs_comp = self.components.get_match('absorbing', False)
+            if not skip_non_abs and len(non_abs_comp) > 0:
+                self.builder.set_non_absorbing_species(pyomo_model, non_abs_comp, check=True)    
+            if hasattr(self,'fixed_params') and len(self.fixed_params) > 0:
+                for param in self.fixed_params:
+                    getattr(pyomo_model, self.__var.model_parameter)[param].fix()
         
-        if not skip_non_abs and len(non_abs_comp) > 0:
-            self.builder.set_non_absorbing_species(self.model, non_abs_comp, check=True)    
-        
-        if hasattr(self,'fixed_params') and len(self.fixed_params) > 0:
-            for param in self.fixed_params:
-                self.model.P[param].fix()
-            
-        return None
+        return pyomo_model
     
     def _add_feed_times(self):
         
@@ -813,89 +840,83 @@ class ReactionModel(WavelengthSelectionMixins):
         """This should try to handle all of the simulation cases"""
     
         # Create the simulator object
-        _print('Making simulator')
         self.create_simulator()
         # Add any previous trajectories, if given
-        _print('Adding traj')
         self._from_trajectories('simulator')
         # Run the simulation
-        _print('Running sim')
         self.run_simulation()
         
         return None
     
     def create_simulator(self):
-        """This should try to handle all of the simulation cases"""
+        """This should try to handle all of the simulation cases
         
-        _print('Setting up simulator:')
+        It should not use the base reaction model! - allows for different
+        discretizations that do not depend on the data
+        
+        """
+        print('Setting up simulation model')
         sim_set_up_options = copy.copy(self.settings.simulator)
-        _print(sim_set_up_options)
-        dis_method = sim_set_up_options.pop('method', 'dae.collocation')
         
+        # Initialization is now defaulted to FESimulator (settings file)
+        dis_method = sim_set_up_options.pop('method', 'fe')
+        
+        # Override the chosen method to fe if dosing or steps are included
         if self._has_step_or_dosing:
             dis_method = 'fe'
         
-        _print(dis_method)
         
-        if dis_method == 'fe':
-            simulation_class = FESimulator
-        else:
-            simulation_class = PyomoSimulator
+        # Choose the proper simulator object
+        simulation_class = PyomoSimulator
+        # if dis_method == 'fe':
+        #     simulation_class = FESimulator
         
-        if self.model is None:
-            # with_data her
-            self.create_pyomo_model(*self.settings.general.simulation_times)
+        self.s_model = self.create_pyomo_model(is_simulation=True)
         
-        self.s_model = self.model.clone()
+        # Fix the optimization variables
+        opt_vars = self.__var.optimization_variables
+        for var in opt_vars:
         
-        if hasattr(self.s_model, self.__var.model_parameter):
-            for param in getattr(self.s_model, self.__var.model_parameter).values():
-                param.fix()
-        
-        _print(simulation_class)
+            if hasattr(self.s_model, var):
+                for param in getattr(self.s_model, var).values():
+                    param.fix()
         
+        # Initialize the simulator instance
         simulator = simulation_class(self.s_model)
-        _print('After class made')
+
+        # Discretize the model
         simulator.apply_discretization(self.settings.collocation.method,
-                                       ncp=self.settings.collocation.ncp,
+                                       ncp=max(2, self.settings.collocation.ncp),
                                        nfe=self.settings.collocation.nfe,
                                        scheme=self.settings.collocation.scheme)
         
-        _print('After disc')
-        
+        # Add the dosing points to the model
         if self._has_step_or_dosing:
             for time in simulator.model.alltime.data():
                 getattr(simulator.model, self.__var.dosing_variable)[time, self.__var.dosing_component].set_value(time)
                 getattr(simulator.model, self.__var.dosing_variable)[time, self.__var.dosing_component].fix()
         
-            #getattr(simulator.model, 'time_step_change')[0].fix()
-        
+        # Add this simulator to the class attributes
         self.simulator = simulator
         print('Finished creating simulator')
         
         return None
-        
+    
+    
     def run_simulation(self):
         """Runs the simulations, may be combined with the above at a later date
         
         """
-        if self._has_dosing_points or isinstance(self.simulator, FESimulator):
+        if isinstance(self.simulator, FESimulator):
             self.call_fe_factory()
         
         simulator_options = self.settings.simulator
         simulator_options.pop('method', None)
-        self.results = self.simulator.run_sim(**simulator_options)
-#        self.results.file_dir = self.settings.general.charts_directory
-    
+        results = self.simulator.run_sim(**simulator_options)
+        self.results_dict['simulator'] = results 
+        self.results = results    
         return None
     
-    # def reduce_spectra_data_set(self, dropout=4):
-    #     """To reduce the computational burden, this can be used to reduce 
-    #     the amount of spectral data used
-        
-    #     """
-    #     A_set = [l for i, l in enumerate(self.model.meas_lambdas) if (i % dropout == 0)]
-    #     return A_set
     
     def bound_profile(self, var, bounds):
         """Wrapper for TemplateBuilder bound_profile method"""
@@ -903,6 +924,7 @@ class ReactionModel(WavelengthSelectionMixins):
         self.builder.bound_profile(var=var, bounds=bounds)
         return None
     
+    
     """Estimators"""
     
     def create_variance_estimator(self, **kwargs):
@@ -930,27 +952,161 @@ class ReactionModel(WavelengthSelectionMixins):
         self.create_estimator(estimator='p_estimator', **kwargs)
         self._from_trajectories('p_estimator')
         return None
+    
+    def _calculate_S_from_Z_data(self):
+        """Calculates the indivdual S profiles from simulated results
+        """
+        
+        C_orig = self.results_dict['simulator'].Z
+        D = self.spectra.data
+    
+        C = interpolate_trajectory(list(D.index), C_orig)
+    
+        non_abs_species = self.components.get_match('absorbing', False)
+        C = C.drop(columns=non_abs_species)
+    
+        indx_list = list(D.index)
+        for i, ind in enumerate(indx_list):
+            indx_list[i] = round(ind, 6)
+        
+        D.index = indx_list
+        
+        assert C.shape[0] == D.values.shape[0]
+        
+        M1 = np.linalg.inv(C.T @ C)
+        M2 = C.T @ D.values
+        S = (M1 @ M2).T
+        S.columns = C.columns
+        S = S.set_index(D.columns)
+        
+        for comp in non_abs_species:
+            S[comp] = 0
+        
+        return S
+    
+    def _check_S_singularity(self, threshold=1e-5):
+        """This is still in development and may not actually be too useful
+        """
+        
+        C_orig = self.results_dict['simulator'].Z
+        D = self.spectra.data
+    
+        C = interpolate_trajectory(list(D.index), C_orig)
+    
+        indx_list = list(D.index)
+        for i, ind in enumerate(indx_list):
+            indx_list[i] = round(ind, 6)
+        
+        D.index = indx_list
+        
+        M = C.T @ C
+        print('This is the determinant of C.T @ C')
+        print(np.linalg.det(M))
+        
+        eigs, U = np.linalg.eigh(M)
+        
+        print('The singular values are:')
+        print(eigs)
+        d = eigs
+        d = np.where(d > 1e-16, d, 1e-16)
+        d = 1/d
+        d = np.where(d > threshold, 0, d)
+        
+        print(d)
+        M1 = np.diag(d)
+        
+        M2 = C.T @ D.values
+        
+        S = (M1 @ M2).T
+        S.columns = C.columns
+        S = S.set_index(D.columns)
+        
+        num_comp = sum(np.where(d > 0, 1, 0))
+
+        return S, num_comp
+        
+    def initialize_S_from_simulation(self, model):
+        """This method takes simulated concentration profiles and the given
+        spectra data to recreate the single species absorbance profiles to be
+        used as initial values in the variance estimation
+        
+        This directly sets the S values in the TemplateBuilder instance. If
+        the S profile is provided, it is fixed in the model.
+        """
+        
+        S = self._calculate_S_from_Z_data()
+        S[S < 0] = 1e-8
+        
+        print('Adding initialized S to the model')
+        
+        for k in S.columns:
+            if self.components[k].absorbing:
+                if self.components[k].S is None:
+                    # Absorbance is taken from simulated values
+                    for l in S.index:        
+                        getattr(model, 'S')[l, k].set_value(float(S[k][l]))
+                else:
+                    # The single species absorbance is provided - fit to the collocation points
+                    for l in S.index:        
+                        S_comp = interpolate_trajectory(list(S.index), self.components[k].S)
+                        S_comp = S_comp.set_index(S.index)
+                        #print(S_comp)
+                        getattr(model, 'S')[l, k].set_value(float(S_comp[k][l]))
+                        getattr(model, 'S')[l, k].fix()
+            else:
+                # Species does not absorb - set to zero
+                for l in S.index:        
+                    getattr(model, 'S')[l, k].set_value(0)
+                    
+        print('Finished initializing the S profiles')        
+        
+        print('Initializing both the C and Z profiles using simulated Z profiles')
+        getattr(self, 'v_estimator').initialize_from_trajectory('C', self.results_dict['simulator'].Z)
+        getattr(self, 'v_estimator').initialize_from_trajectory('Z', self.results_dict['simulator'].Z)
+        
+        return None
         
     def initialize_from_simulation(self, estimator='p_estimator'):
+        """This method initializes the model using simulated data
         
-        if not hasattr(self, 's_model'):
-            _print('Starting simulation for initialization')
+        """
+        if not hasattr(self, 's_model') or self.s_model is None:
             self.simulate()
-            _print('Finished simulation, updating variables...')
 
-        _print(f'The model has the following variables:\n{get_vars(self.s_model)}')
         vars_to_init = get_vars(self.s_model)
         
-        _print(f'The vars_to_init: {vars_to_init}')
+        #_print(f'The vars_to_init: {vars_to_init}')
         for var in vars_to_init:
-            if hasattr(self.results, var):    
-                _print(f'Updating variable: {var}')
+            if hasattr(self.results, var) and var != 'S':
+                #_print(f'Updating variable: {var}')
+                getattr(self, estimator).initialize_from_trajectory(var, getattr(self.results, var))
+            elif var == 'S' and hasattr(self.results, 'S'):
                 getattr(self, estimator).initialize_from_trajectory(var, getattr(self.results, var))
             else:
                 continue
         
         return None
     
+    # def initialize_from_reduced_spectral_data(self, estimator='p_estimator'):
+        
+    #     if not hasattr(self, 's_model'):
+    #         _print('Starting simulation for initialization')
+    #         self.simulate()
+    #         _print('Finished simulation, updating variables...')
+
+    #     _print(f'The model has the following variables:\n{get_vars(self.s_model)}')
+    #     vars_to_init = get_vars(self.s_model)
+        
+    #     _print(f'The vars_to_init: {vars_to_init}')
+    #     for var in vars_to_init:
+    #         if hasattr(self.results, var):    
+    #             _print(f'Updating variable: {var}')
+    #             getattr(self, estimator).initialize_from_trajectory(var, getattr(self.results, var))
+    #         else:
+    #             continue
+        
+    #     return None
+    
     def create_estimator(self, estimator=None):
         """This function handles creating the Estimator object
         
@@ -972,7 +1128,7 @@ class ReactionModel(WavelengthSelectionMixins):
         else:
             raise ValueError('Keyword argument estimator must be p_estimator or v_estimator.')  
         
-        model_to_clone = self.model
+        model_to_clone = self.create_pyomo_model()
         
         setattr(self, f'{estimator[0]}_model', model_to_clone.clone())
         setattr(self, estimator, Estimator(getattr(self, f'{estimator[0]}_model')))
@@ -980,12 +1136,20 @@ class ReactionModel(WavelengthSelectionMixins):
                                                       ncp=self.settings.collocation.ncp,
                                                       nfe=self.settings.collocation.nfe,
                                                       scheme=self.settings.collocation.scheme)
-        _print('Starting from_traj')
+     
         self._from_trajectories(estimator)
         
-        _print('Starting sim init')
-        if self.settings.parameter_estimator.sim_init and estimator == 'p_estimator':
+        if self.settings.parameter_estimator.sim_init and estimator == 'v_estimator':
             self.initialize_from_simulation(estimator=estimator)
+            # This really doesn't seem to help as much - use the simpler implementation above
+            #self.initialize_S_from_simulation(getattr(self, 'v_model'))
+            
+        # What is S between VE and PE?
+        if self.settings.parameter_estimator.sim_init and estimator == 'p_estimator':
+            if hasattr(self, 'v_estimator'):
+                self.initialize_from_variance_trajectory()
+            else:
+                self.initialize_from_simulation(estimator=estimator)
         
         if self._has_step_or_dosing:
             for time in getattr(self, estimator).model.alltime.data():
@@ -1074,17 +1238,17 @@ class ReactionModel(WavelengthSelectionMixins):
         remove the VarianceEstimator being required to be implemented by the user
         
         """
-        _print('Starting the RunOpt method')
+        print(f'*** KIPET version {version_number}')
+        print(f'*** Starting the parameter estimation procedure')
         
-        # Make the model if not present
-        if self.model is None:    
-            self.create_pyomo_model()  
-            _print('Generating model')
+        print(f'\n*** Simulating the model with initial values')
+        self.simulate()
+        print(f'*** Simulation completed succesfully')
         
         # Check if all needed data for optimization available
         if not self.allow_optimization:
             raise ValueError('The model is incomplete for parameter optimization')
-            
+        
         # Some settings are required together, this method checks this
         self._update_related_settings()
         
@@ -1103,6 +1267,7 @@ class ReactionModel(WavelengthSelectionMixins):
             """If the data is spectral and not all variances are provided, VE needs to be run"""
             
             # Create the VE
+            print('*** Generating VarianceEsitmator Instance')
             self.create_estimator(estimator='v_estimator')
             self.settings.variance_estimator.solver_opts = self.settings.solver
             # Optional max device variance
@@ -1113,6 +1278,7 @@ class ReactionModel(WavelengthSelectionMixins):
             #     variances_with_delta = self.solve_variance_given_delta()
 
             else:
+                print('*** Starting the variance estimator')
                 self.run_ve_opt()
                 
                 #print('Finished VE Opt - moving to PE opt')
@@ -1126,7 +1292,7 @@ class ReactionModel(WavelengthSelectionMixins):
                     comp.variance = 1
                 
         # Create ParameterEstimator
-        _print('Making PEstimator')
+        print('*** Generating ParameterEsitmator Instance')
         self.create_estimator(estimator='p_estimator')
         
         variances = self.components.variances
@@ -1165,6 +1331,7 @@ class ReactionModel(WavelengthSelectionMixins):
         self.settings.parameter_estimator.variances = self.variances
         
         # Run the PE
+        print('*** Solving the parameter fitting problem...\n')
         self.run_pe_opt()
         
         # Save results in the results_dict
@@ -1233,6 +1400,8 @@ class ReactionModel(WavelengthSelectionMixins):
         estimator = getattr(self, obj)
         method = getattr(estimator, f'{category}_from_trajectory')
         
+        print(variable)
+        
         if variable is None:
             vars_to_init = get_vars(estimator.model)
         
@@ -1349,7 +1518,7 @@ class ReactionModel(WavelengthSelectionMixins):
         
         """
         if self.model is None:
-            self.create_pyomo_model()
+            self.model = self.create_pyomo_model()
             
         kwargs = {}
         kwargs['solver_opts'] = self.settings.solver
@@ -1417,6 +1586,10 @@ class ReactionModel(WavelengthSelectionMixins):
     def unwanted_contribution(self, variant, **kwargs):
         
         self._G_contribution = variant
+        
+        if 'St' not in kwargs:
+            kwargs['St'] = self.build_stoich_matrix(as_dict=True)
+        
         self._G_data = {'G_contribution': variant,
                         'Z_in': kwargs.get('Z_in', dict()),
                         'St': kwargs.get('St', dict()),
@@ -1480,6 +1653,150 @@ class ReactionModel(WavelengthSelectionMixins):
         
     """MODEL FUNCTION AREA"""
     
+    @staticmethod
+    def _set_up_stoich_mat(rm, St, reaction, component, value):
+    
+        if component in rm.components.names and reaction in rm.algs_dict:
+            St[reaction][rm.components.names.index(component)] = value
+        else:
+            pass
+        
+    
+    def _create_stoich_dataframe(self):
+        """Builds the dataframe used to hold the stoichiometric matrix
+        """
+            
+        if self.algebraics.get_match('is_reaction', True) is None:
+            raise ValueError('You need to declare reaction expressions')
+        
+        odes = self.odes_dict
+        reaction_exprs = self.algebraics.get_match('is_reaction', True)
+        dr = pd.DataFrame(np.zeros((len(self.components), len(reaction_exprs))), columns=reaction_exprs, index=self.components.names)    
+           
+        return dr
+    
+    
+    def build_from_reaction_matrix(self, St):
+        """The user can provide a stoichiometric matrix and have the reaction
+        ODEs generated from it
+        
+        How should St be input? As a dict with lists? Yes for now
+        """
+        dr = self._create_stoich_dataframe()
+        comps = self.components.names
+        
+        for rxn, s_list in St.items():
+            dr.loc[:, rxn] = s_list
+
+        # Now we have the dr dataframe, put together the reactions
+        for comp in comps:
+            ode = 0
+            for rxn in dr.columns:
+                ode += dr.loc[comp, rxn]*self.algs_dict[rxn].expression
+                
+            self.add_ode(comp, ode)
+
+        return None
+
+    def build_stoich_matrix(self, as_dict=False):
+        """Generate a dictionary or dataframe representing the stoichiometric
+        reaction matrix given the reaction expressions and the ODES are
+        defined
+        
+        Debugging code still left here because this may still lead to errors
+        """
+        from pyomo.core.expr.numeric_expr import DivisionExpression, ProductExpression, NegationExpression, SumExpression
+      
+        if self.odes_dict is None or len(self.odes_dict) == 0:
+            raise ValueError('You need to input the reaction ODEs')
+          
+        dr = self._create_stoich_dataframe()
+        all_rxns = {key: self.algs_dict[key].expression for key in self.algebraics.get_match('is_reaction', True)}
+        
+        def check_expr_type(expr):
+        
+            if isinstance(expr, NegationExpression):
+                expr_use = expr.args[0]
+                scalar = -1       
+                return expr_use, -1
+            else:
+                return expr, 1
+    
+        for comp, ode in self.odes_dict.items():
+    
+        
+            expr = ode.expression
+            #print('')
+            #print(f'Looking at {comp} and {expr}')
+            #print(comp, expr)
+            #expr_new = 0
+            
+            #print(f'The ODE expression being handled is:\n     {expr}\n')
+            #print(f'The type of expression being handled is:\n     {type(expr)}\n')
+        
+            #print(type(expr))
+        
+            expr_use = expr
+        
+            scalar = 1
+        
+            if isinstance(expr, NegationExpression):
+                expr_use = expr.args[0]
+                scalar = -1        
+                
+            if isinstance(ode.expression, (DivisionExpression, ProductExpression)):
+            #    _print(f'The number of terms in this expression is: 1\n')
+                expr_use = expr 
+            
+            if isinstance(ode.expression, SumExpression):
+                expr_use = expr.args
+                
+            #    term = self.check_term(expr, convert_to)
+            #    expr_new = scalar*term
+            coeff = 1
+            
+            #else:
+            #print(f'The number of terms in this expression is: {len(expr.args)}\n')
+                
+            expr_use = [expr_use]
+ 
+ #           print('\nStarting For LOOP')
+            for i, term in enumerate(expr_use):
+  #              print(i, term)
+   #             print('')
+    #            print(type(term))
+                
+                if not isinstance(term, list):
+                    term = [term]
+                for t in term:
+     #               print(t)
+                    t, scalar_update = check_expr_type(t)
+                    
+                    for name, rxn in all_rxns.items():
+      #                  print(name, rxn)
+                        if t is rxn:
+       #                     print(f'{name}: {t} == {rxn}')
+                            dr.loc[comp, name] = scalar*scalar_update
+                        #print(f'Equal to {name}: {term == rxn}')
+                
+            #term = self.check_term(term, convert_to)
+            #expr_new += scalar*term
+        
+        #_print('The new expression is:\n')
+        #_print(f'     {expr_new}')
+            
+        #self.expression = expr_new
+        #self.units = getattr(pyo_units, convert_to)
+        self.St = dr
+        
+        if as_dict:
+            _dict = dict(dr)
+            St = {k: list(v.values) for k, v in _dict.items()}
+            return St
+            
+        else:
+            return dr
+        
     def make_reaction_table(self, stoich_coeff, rxns):
         
         df = pd.DataFrame()
@@ -1528,7 +1845,7 @@ class ReactionModel(WavelengthSelectionMixins):
                 output += f'{model} False\n'
                 output_dict[model] = False
             
-        print(output)
+        #print(output)
         return output_dict
     
     @property
@@ -1738,18 +2055,22 @@ class ReactionModel(WavelengthSelectionMixins):
         
 
     def ae(self, alg_var):
+        """Quick method to get the algebraic expression for alg_var"""
         return self.algs_dict[alg_var].expression
     
     def ode(self, ode_var):
+        """Quick method to get the algebraic expression for ode_var"""
         return self.odes_dict[ode_var].expression
 
     def get_state(self, comp):
+        """Quick method to get the component or state variable"""
         if comp in self.components:
             return self.components[comp]
         elif comp in self.states:
             return self.states[comp]
  
     def get_alg(self, comp):
+        """Quick method to get the algebraic variable"""
         if comp in self.algebraics:
             return self.algebraics[comp]
         
